Unit TEvoCreatureSrc;
Interface
Uses
	Classes,
        EvoStoreSrc,
        graphics,
        windows,
        cybmanip,
        math,
	sysutils;

type
TEvoCreature = class 		 {TEvoCreature}
        FoundPartner : Boolean;
        FoundFood : Boolean;
	Sex : TSexes;
        AnimalState : TAnimalStates;
	LocPoint : TPoint;
        Energy : Integer;
        Heading : Integer;
        StepSize : Integer;
        CreatureStorePtr : PCreatureProperties;
        LastFoodPtr : TEvoCreature;
        FoodPtr : TEvoCreature;
        BreedPtr : TEvoCreature;
        LastBreedPtr : TEvoCreature;
        EatingCreature : TEvoCreature;

	constructor Create;
	procedure Paint(aCanvas : TCanvas);
	function Breed( aEvoCreature : TEvoCreature) : boolean;
	function Encounter(aEvoCreature : TEvoCreature) : Integer;
	procedure Eats(aEvoCreature : TEvoCreature);
	procedure Competes(aEvoCreature : TEvoCreature);
	function CanInteract( aEvoCreature : TEvoCreature) : boolean;
        function IsMateAble(aEvoCreature : TEvoCreature) : boolean;
        function IsSameGene(aEvoCreature : TEvoCreature) : boolean;
        function WillEat(aEvoCreature : TEvoCreature) : integer;
        function RandomSex : Tsexes;
        function IsAnimal : Boolean;
        function IsAvailable : Boolean;
        procedure CheckIfDead;
        procedure ResetFields;
        procedure moveDelta;
	procedure Influence( var arect : TRect);
	procedure SetFromRecord(ApCreatureProperties : tCreatureProperties);
	procedure Step;
	procedure AddToCensus;
        function Alive : Boolean;
        procedure DecideNextHeading;
        procedure RandomHeading;
end; {TEvoCreature}

var
   nexTEvoCreature : TEvoCreature;

Function getnewEvoCreature : TEvoCreature;
procedure ClearReincarnationList;

implementation

Function getnewEvoCreature : TEvoCreature;

begin
   if (ReincarnationList.count > 0) then begin
       result := ReincarnationList[0];
       with result do begin
         resetfields;
         Energy := 0;
         StepSize := 0;
         CreatureStorePtr := nil;
       end;
       ReincarnationList.delete(0);
   end
   else begin
       result := TEvoCreature.create;
   end;
end;

procedure ClearReincarnationList;
var
   EvoCreature : TEvoCreature;
   lc, mc : integer;

begin
   mc := pred(ReincarnationList.count);
   for lc := 0 to mc do begin
	EvoCreature := TEvoCreature(ReincarnationList[lc]);
        EvoCreature.free;
   end;
   ReincarnationList.clear;
end;

constructor TEvoCreature.Create;

begin
  inherited create;
  LocPoint := Point( 0, 0);
  ResetFields;
end;

procedure TEvoCreature.AddToCensus;
var
  gs : integer;

begin
  inc(paintcount);
  gs := CreatureStorePtr^.Geneset;
  CreatureCensus^[gs] := CreatureCensus^[gs]+1;
  if (Sex = esMale) then MaleCreatureCensus[gs] := MaleCreatureCensus[gs]+1
                    else FeMaleCreatureCensus[gs] := FeMaleCreatureCensus[gs]+1;
end;

procedure TEvoCreature.Paint(aCanvas : TCanvas);
var
  hx, hy, rx, ry, tx, ty, lx, ly : integer;

begin
   if (AnimalState <> taDead) then begin
      with aCanvas, locpoint do begin
         lx := x*dx;
         ly := y*dy;
         rx := (lx+dx);
         ry := (ly+dy);
         ty := ly-dx2;
         tx := lx+dx2;
         CopyRect(rect(lx, ly, rx, ry), CreatureStorePtr^.graphic.canvas, stdrect);
         if CreatureStorePtr^.Showenergy then textout(tx, ty, inttostr(self.Energy));
         if CreatureStorePtr^.Showsex then MyDraw(aCanvas,lx, ty, SexPics[Sex]);
         if (CreatureStorePtr^.Showactivity and (AnimalState <> taIdle)) then MyDraw(aCanvas, lx, ry, AnimalStatePics[AnimalState]);
         if (EatingCreature <> nil) and CreatureStorePtr^.Showpartner then MyDraw(aCanvas,tx, ry, EatingCreature.CreatureStorePtr^.graphic);
         AddToCensus;
      end;
   end;
end;

function TEvoCreature.RandomSex : Tsexes;
Var
   dif, tot, td,td1,randt : double;

begin
    td := MaleCreatureCensus[CreatureStorePtr^.Geneset] + 1;
    td1 := FeMaleCreatureCensus[CreatureStorePtr^.Geneset] + 1;
    tot := (td+td1);
    dif := (td-td1);
    randt := 0.5 + (dif / tot);
    td := random;
    if (td > randt) then result := esMale
                    else result := esFemale;
end;

procedure TEvoCreature.RandomHeading;

begin
   Heading := random(pi360);
end;

procedure TEvoCreature.SetFromRecord(ApCreatureProperties : tCreatureProperties);

begin
        LocPoint := point(trunc(random(mx)), trunc(random(my)));
        Energy := 2*trunc(random(ApCreatureProperties.Energy));
        if (ApCreatureProperties.StepSize > 0) then begin
          StepSize := 2*trunc(random(ApCreatureProperties.StepSize));
          StepSize := greatint(stepsize, 1);
          RandomHeading;
        end
        else begin
          StepSize := trunc(random(abs(ApCreatureProperties.StepSize)));
          StepSize := -2*StepSize;
          Heading := 0;
        end;
        CreatureStorePtr := @CreatureStores[ApCreatureProperties.Geneset];
        Sex := RandomSex;
end;

function TEvoCreature.IsAvailable : Boolean;

begin
  result := (AnimalState in [taIdle, taBeingEaten, taCompeting]);
end;

function TEvoCreature.IsSameGene(aEvoCreature : TEvoCreature) : boolean;

begin
  result := (CreatureStorePtr = aEvoCreature.CreatureStorePtr);
end;

function TEvoCreature.IsAnimal : Boolean;

begin
  result := (CreatureStorePtr^.stepsize > 0);
end;

function TEvoCreature.IsMateAble(aEvoCreature : TEvoCreature) : boolean;
Var
  bi, ism, isf : boolean;

begin
  result := IsSameGene( aEvoCreature);
  if result then begin
      ism :=  ((Sex = esMale) and (aEvoCreature.sex = esFemale));
      isf :=  ((Sex = esFeMale) and (aEvoCreature.sex = esMale));
      bi := (AnimalState <> taBreeding) and (aEvoCreature.AnimalState <> taBreeding);
      result := ((ism or isf) and bi);
  end;
end;

function TEvoCreature.Breed(aEvoCreature : TEvoCreature) : boolean;
Var
   tb1, tb : boolean;
   dble, el, lc : integer;
   CEvoCreature : TEvoCreature;

begin
  result := IsMateAble(aEvoCreature);
  if result then begin
    tb1 := IsAnimal;
    if tb1 then result := CanInteract( aEvoCreature);
    if result then begin
      lc := 1;
      dble := 2*CreatureStorePtr^.Energy;
      tb := ((lc <= CreatureStorePtr^.NumberChildren) and (energy > dble) and (aEvoCreature.energy > dble));
      if tb then begin
          AnimalState := taBreeding;
          aEvoCreature.AnimalState := taBreeding;
          heading := 0;
          aEvoCreature.heading := 0;
          LastFoodPtr := nil;
          LastBreedPtr := nil;
      end
      else begin
        result := false;
        exit;
      end;
      while tb do begin
        CEvoCreature := getnewEvoCreature;
        CEvoCreature.SetFromRecord(CreatureStorePtr^);
        if tb1 then CEvoCreature.locpoint := locpoint;
        BreadCreatureList.add(CEvoCreature);
        el := CEvoCreature.Energy div 2;
        Energy := Energy - el;
        aEvoCreature.Energy := aEvoCreature.Energy - el;
        inc(lc);
        tb := ((lc <= CreatureStorePtr^.NumberChildren) and (energy > dble) and (aEvoCreature.energy > dble));
      end;
    end
    else if tb1 then begin
         FoundPartner := (LastBreedPtr = aEvoCreature);
         if foundpartner then begin
           BreedPtr := aEvoCreature;
           aEvoCreature.FoundPartner := true;
           aEvoCreature.BreedPtr := self;
         end;
         nexTEvoCreature := aEvoCreature;
    end;
  end;
end;

procedure TEvoCreature.ResetFields;

begin
    AnimalState := taIdle;
    FoundPartner := false;
    FoundFood := false;
    Foodptr := nil;
    BreedPtr := nil;
    EatingCreature := nil;
end;

function TEvoCreature.Alive : Boolean;

begin
   result := (energy > CreatureStorePtr^.minenergy) and (stepsize <> 0);
   if result and isanimal then result := result and (stepsize > 0);
end;

function TEvoCreature.WillEat(aEvoCreature : TEvoCreature) : integer;

begin
   if isanimal then result := pos(aEvoCreature.CreatureStorePtr^.DisplayName, CreatureStorePtr^.Foodlist)
               else result := 0;
end;

function TEvoCreature.Encounter(aEvoCreature : TEvoCreature) : Integer;

begin
    result := 0;
    if (aEvoCreature.animalstate = tadead) then exit;
    result := -1;
    if IsSameGene(aEvoCreature) then exit;
    result := WillEat(aEvoCreature);
    if (result > 0) then begin
      if CanInteract( aEvoCreature) then begin
        result := 1;
      end
      else begin
         FoundFood := (Foodptr = Lastfoodptr);
         if FoundFood then Foodptr := aEvoCreature;
      end;
    end;
end;

procedure TEvoCreature.moveDelta;
var
  endpoint : Tpoint;
  ang : double;
  lc, mc : integer;

  procedure calcnewpos;
  begin
      ang := Heading / 360;
      lc := round(StepSize * cos(ang));
      mc := round(StepSize * sin(ang));
  end;

begin
    calcnewpos;
    if ((lc = 0) and (mc = 0)) then begin
      RandomHeading;
      StepSize := StepSize + 1;
      calcnewpos;
    end;
    with endpoint do begin
      x := LocPoint.x + lc;
      y := LocPoint.y + mc;
      if (x > mx ) then x := 0
                   else if (x < 0) then x := mx;
      if (y > my) then y := 0
                  else if (y < 0) then y := my;
    end;
    energy := energy - stepsize;
    locpoint := endpoint;
end;

procedure TEvoCreature.Step;
Var
  tb1, tb : boolean;
  aEvoCreature : TEvoCreature;

  Procedure NewDelta;
  begin
    RandomHeading;
    moveDelta;
  end;

begin
  tb1 := (animalstate = taEating);
  if tb1 then begin
    aEvoCreature := EatingCreature;
    tb1 := ((energy < (CreatureStorePtr^.maxenergy - stepsize))
           and (EatingCreature.Animalstate <> taDead)
           and (EatingCreature.energy > (EatingCreature.CreatureStorePtr^.minenergy + EatingCreature.stepsize))
           and caninteract(EatingCreature));
  end;
  ResetFields;
  if ISAnimal then begin
    if tb1 then begin
       EatingCreature := aEvoCreature;
       animalstate := taEating;
    end
    else begin
      if (heading = 0) then begin
          tb := (energy >= CreatureStorePtr^.maxenergy);
          if tb then NewDelta
          else begin
              heading := heading - 1;
              if (heading < -1) then NewDelta;
          end;
      end
      else moveDelta;
    end;
  end
  else energy := energy - stepsize;
  energy := lessint(energy, CreatureStorePtr^.maxenergy);
end;

function TEvoCreature.CanInteract( aEvoCreature : TEvoCreature) : boolean;
var
  arect : trect;

begin
   arect := rect((LocPoint.x-1),(LocPoint.y-1),(LocPoint.x+1),(LocPoint.y+1));
   result := myPtInRect(arect, aEvoCreature.locpoint) and (aEvoCreature.AnimalState <> taDead);
end;

procedure TEvoCreature.DecideNextHeading;
var
   aEvoCreature : TEvoCreature;
   ti, dx, dy : integer;
   aext : extended;

begin
  if IsAnimal then begin
     aEvoCreature := nil;
     if FoundPartner then begin
         aEvoCreature := BreedPtr;
         LastBreedPtr := BreedPtr;
         LastFoodPtr := nil;
     end
     else if FoundFood then begin
         aEvoCreature := FoodPtr;
         LastFoodPtr := FoodPtr;
         LastBreedPtr := nil;
     end;
     if (aEvoCreature <> nil) then begin
        dx := aEvoCreature.locpoint.x - locpoint.x;
        dy := aEvoCreature.locpoint.y - locpoint.y;
        if (dx = 0) then heading := pi360
        else begin
           aext := ArcTan2(dy, dx);
           ti := round(aext*360);
           if (ti < 0) then heading := pi360-ti
                       else heading := ti;
        end;
     end;
  end;
end;

procedure TEvoCreature.Influence( var arect : TRect);
Var
   lc : integer;

begin
   lc := CreatureStorePtr^.InFluenceDistance;
   with arect, LocPoint do begin
        Left := greatint((x-lc), 0);
       right := lessint((x+lc), mx);
         top := greatint((y-lc), 0);
      bottom := lessint((y+lc), my);
   end;
end;
procedure TEvoCreature.CheckIfDead;

begin
    if not(alive) then begin
      if (Animalstate = taCompeting) then inc(deadcount)
      else if (Animalstate = taBeingEaten) then inc(eatencount);
      Animalstate := taDead;
    end;
end;

procedure TEvoCreature.Eats(aEvoCreature : TEvoCreature);

begin
   heading := 0;
   AnimalState := taEating;
   if aEvoCreature.IsAnimal then begin
      energy := energy + aEvoCreature.energy;
      aEvoCreature.energy := 0;
   end
   else begin
     energy := energy + Stepsize;
     aEvoCreature.energy := aEvoCreature.energy -  Stepsize;
     aEvoCreature.AnimalState := taBeingEaten;
   end;
   EatingCreature := aEvoCreature;
   aEvoCreature.EatingCreature := self;
   aEvoCreature.CheckIfDead;
   if (aEvoCreature.Animalstate = taDead) then RandomHeading;
   LastFoodPtr := nil;
end;

procedure TEvoCreature.Competes(aEvoCreature : TEvoCreature);
Var
   lc, mc : integer;

begin
   if not(CanInteract( aEvoCreature)) then exit;
   if IsAnimal then begin
     lc := stepsize;
     mc := aEvoCreature.stepsize;
     RandomHeading;
     aEvoCreature.RandomHeading;
   end
   else begin
      lc := -1*stepsize;
      mc := -1*aEvoCreature.stepsize;
   end;
   AnimalState := taCompeting;
   aEvoCreature.AnimalState := taCompeting;
   energy := energy - mc;
   aEvoCreature.energy := aEvoCreature.energy -  lc;
   CheckIfDead;
   aEvoCreature.CheckIfDead;
end;

{ End of code  }
end.

